package fr.unice.javinspector.GUI.util;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.Hashtable;
import java.util.ArrayList;

/**
 *
 * <p>Title: Class JTabbedPaneWithCloseIcons</p>
 * <p>Description: This class created a tabbed pane with an X icon for each component that
 * allow the user to directly close them.</p>
 * <p>Company: UNSA</p>
 * @author Charbit Lucas - Gahide Jrome - Galbois Xavier - Zerbib Gabriel
 * @version 1.0
 */

public class JTabbedPaneWithCloseIcons extends JTabbedPane implements MouseListener {
  private Hashtable names = new Hashtable();
  private BeforeCloseEvent beforeCLoseEvent = null;
  private boolean closeLastTab = true;

  /**
   * Default constructor
   */
  public JTabbedPaneWithCloseIcons() {
    this(true);
  }

  /**
   * Create a closeable tabbedPane
   * @param closeLastTab : specify if the first component of the tabbedPane is closeable too.
   */
  public JTabbedPaneWithCloseIcons(boolean closeLastTab){
    super();
    addMouseListener(this);
    this.closeLastTab = closeLastTab;
  }

  /**
   * Add a component to the tabbed pane
   * @param title : the tittle of the component
   * @param component : the component to add
   */
  public void addTab(String title, Component component) {
      this.addTab(title, component, null, null);
  }

  /**
   * Add a component to the tabbed pane
   * @param title : the tittle of the component
   * @param component : the component to add
   * @param tooltiptext : the tooltip text associated to thos component
   */
  public void addTab(String title, Component component, String tooltiptext) {
      this.addTab(title, component, null, tooltiptext);
  }

  /**
   * Add a component to the tabbed pane
   * @param title : the tittle of the component
   * @param component : the component to add
   * @param tooltiptext : the tooltip text associated to thos component
   * @param extraIcon : a second a icon to display near the X icon
   */
  public void addTab(String title, Component component, Icon extraIcon, String tooltiptext) {
    if(!names.containsKey(title)) {
      names.put(title, title);
      if (getTabCount() == 0 && !closeLastTab) {
        super.addTab(title, extraIcon, component, tooltiptext);
      }
      else {
        super.addTab(title, new CloseTabIcon(extraIcon), component, tooltiptext);
      }
    }
  }

  /**
   * This function add a BeforeCloseEvent to the tabbedPane
   * @param l : the before close event
   */
  public void addBeforeCloseListener(BeforeCloseListener l) {
      listenerList.add(BeforeCloseListener.class, l);
  }

  /**
   * Fire an event when a component in the tabbed pane is going to be close
   */
  protected void fireBeforeClose() {
    // Guaranteed to return a non-null array
    Object[] listeners = listenerList.getListenerList();
    // Process the listeners last to first, notifying
    // those that are interested in this event
    for (int i = listeners.length-2; i>=0; i-=2) {
        if (listeners[i]==BeforeCloseListener.class) {
            // Lazily create the event:
            if (beforeCLoseEvent == null)
                beforeCLoseEvent = new BeforeCloseEvent(this);
            ((BeforeCloseListener)listeners[i+1]).beforeClose(beforeCLoseEvent);
        }
    }
}

  /**
   * Needed by the interface MouseListened
   * This function is called when a mouse clicked event occured
   * @param e : the mouse event
   */
  public void mouseClicked(MouseEvent e) {
    int tabNumber=getUI().tabForCoordinate(this, e.getX(), e.getY());
    if (tabNumber < 0) return;
    if (tabNumber != 0 || closeLastTab) {
      Rectangle rect=((CloseTabIcon)getIconAt(tabNumber)).getBounds();
      if (rect.contains(e.getX(), e.getY())) {
        //the tab is being closed
        fireBeforeClose();
        names.remove(this.getTitleAt(tabNumber));
        this.removeTabAt(tabNumber);
        fireStateChanged();
      }
    }
  }

  /**
   * Needed by the MouseListened interface
   * @param e
   */
  public void mouseEntered(MouseEvent e) {}
  public void mouseExited(MouseEvent e) {}
  public void mousePressed(MouseEvent e) {}
  public void mouseReleased(MouseEvent e) {}


}
/**
 * The class which generates the X icon for the tabs. The constructor
 * accepts an icon which is extra to the X icon.
 */
class CloseTabIcon implements Icon {
  private int x_pos;
  private int y_pos;
  private int width;
  private int height;
  private Icon fileIcon;

  public CloseTabIcon(Icon fileIcon) {
    this.fileIcon=fileIcon;
    width=16;
    height=16;
  }

  public void paintIcon(Component c, Graphics g, int x, int y) {
    this.x_pos=x;
    this.y_pos=y;

    Color col=g.getColor();

    g.setColor(Color.black);
    int y_p=y+2;
    g.drawLine(x+1, y_p, x+12, y_p);
    g.drawLine(x+1, y_p+13, x+12, y_p+13);
    g.drawLine(x, y_p+1, x, y_p+12);
    g.drawLine(x+13, y_p+1, x+13, y_p+12);
    g.drawLine(x+3, y_p+3, x+10, y_p+10);
    g.drawLine(x+3, y_p+4, x+9, y_p+10);
    g.drawLine(x+4, y_p+3, x+10, y_p+9);
    g.drawLine(x+10, y_p+3, x+3, y_p+10);
    g.drawLine(x+10, y_p+4, x+4, y_p+10);
    g.drawLine(x+9, y_p+3, x+3, y_p+9);
    g.setColor(col);
    if (fileIcon != null) {
      fileIcon.paintIcon(c, g, x+width, y_p);
    }
  }

  public int getIconWidth() {
    return width + (fileIcon != null? fileIcon.getIconWidth() : 0);
  }

  public int getIconHeight() {
    return height;
  }

  public Rectangle getBounds() {
    return new Rectangle(x_pos, y_pos, width, height);
  }
}
