package fr.unice.javinspector.core.engineSearch.advancedEngineSearch;

import java.io.*;
import java.util.*;
import javax.swing.*;
import java.lang.reflect.*;
import fr.unice.javinspector.util.*;
import fr.unice.javinspector.core.analyser.*;
import fr.unice.javinspector.core.engineSearch.*;
import fr.unice.javinspector.core.engineSearch.advancedEngineSearch.parser.*;



/**
 * <p>Title: AdvancedEngineSearch.</p>
 * <p>Description: This class gives an advanced Search. This search is based on a
 * expression ( a string ). The expression has to be parse by a parser.
 * The grammar is describe in the AdvSearchParser class documentation.
 * </p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: UNICE</p>
 * @author Zerbib Gabriel - Charbit Lucas - Galbois Xavier - Gahide Jrme
 * @version 1.0
 * @see AdvSearchParser
 */

public class AdvancedEngineSearch {

  private Hashtable criterionList;

  private Inspector inspector = null;

  private ClassifierAnalyser classifier = null;


  private AdvSearchParser parser;

  /**
   * TYPE_OF_SEARCH = -1 by default
   * TYPE_OF_SEARCH = 0 on a inspector search
   * TYPE_OF_SEARCH = 1 on a ClassifierAnalyser search
   * @see ClassifierAnalyser
   * @see Inspector
   */
  private int TYPE_OF_SEARCH = -1;

  private Tree exprTree = new Tree();

  private String toto = "";

  /**
   * Construct a new Advanced search with the inspector
   * @param inspector the inspector to use for making some search.
   * @see Inspector
   */
  public AdvancedEngineSearch(Inspector inspector) {
    this.inspector = inspector;
    criterionList = new Hashtable();
    TYPE_OF_SEARCH = 0;
    parser = new AdvSearchParser(new ByteArrayInputStream(toto.getBytes()));
  }

  /**
   * Construct a new Advanced search with the classifier
   * @param classifier the classifier to use for making some search.
   * @see ClassifierAnalyser
   */
  public AdvancedEngineSearch(ClassifierAnalyser classifier) {
    this.classifier = classifier;
    criterionList = new Hashtable();
    TYPE_OF_SEARCH = 1;
    parser = new AdvSearchParser(new ByteArrayInputStream(toto.getBytes()));
  }

  /**
    * To change the inspector if necessary
    * @param inspector the new inspector
    */
   public void setInspector(Inspector inspector) {
     this.inspector = inspector;
     TYPE_OF_SEARCH = 0;
   }

   /**
    * To change the classifier analyser if necessary.
    * @param ca the new classifier analyser
    */
   public void setClassifierAnalyser(ClassifierAnalyser ca) {
     this.classifier = ca;
     TYPE_OF_SEARCH = 1;
  }


  /**
   * Add a criterion to the search
   * If a problem occur, the criterion is not added
   * and a message is printed on std err.
   * @param c the criterion to add.
   */
  public void addCriterion(Criterion c){
    try {
      // neccsity to put to yes if we want
      // the c.isCriterionValid() to work with this search
      c.setUseCriterion(EngineSearchConst.YES);
    }
    catch (SearchEngineException ex) {
      // ex.printStackTrace();
      System.err.println(ex.getMessage());
      System.err.println("Error in addCriterion");
    }
    criterionList.put(c.getName(),c);
  }

  /**
   * This HashTable represent the list of all criterion that are
   * present in this advanced search.
   * @return the Hashtable of the criterion.
   */
  public Hashtable getCriterion(){
    return criterionList;
  }

  /**
   * Check if the criterion c is present in the Advanced Search.
   * @param c the criterion to check.
   * @return true if the criterion is in the search
   */
  public boolean checkCriterion(Criterion c){
    return criterionList.get(c.getName()) != null;
  }

  /**
   * Check if the criterion name is present in the Advanced Search.
   * @param criterionName the criterionName to check.
   * @return true if the criterion is in the search
   */
  public boolean checkCriterion(String criterionName){
   return criterionList.get(criterionName) != null;
 }

 /**
  * Set the Tree of expression for this advenced Search.
  * This procedure automaticaly check for some error in the
  * tree. If the tree is not good sementicaly, an exception
  * will be throw.
  * @param t the tree to set.
  * @throws Exception
  * @see Tree
  */
 public void setExpressionTree(Tree t)
     throws Exception{
   if(is_ValidateTree(t))
     this.exprTree = t;
   else
     throw new Exception("The tree in setExpressionTree is not valid..!!");
 }

  /**
   * Set the expression to parse.
   * This procedure do the parsing of the String and assign
   * the tree. if the String is not Valid, an analys execption is thrown.
   * It also check if the tree is a Valid tree (sementicaly).
   * If an error occur, an error message will be printed on the std err.
   * @param expr the expression to parse.
   * @throws ParseException
   */
  public void setExpression(String expr)
  throws ParseException{
    Tree tmpTreeAnalysed = new Tree();

    if( expr != null && expr.compareTo("") != 0){
      ByteArrayInputStream bais = new ByteArrayInputStream(expr.getBytes());


      parser.ReInit(bais);
//      try {
        tmpTreeAnalysed = parser.Start();
//      }
//      catch (ParseException ex) {
//        System.err.println("Erreur durant l'analyse de l'expression");
//        System.err.println(ex.getMessage());
//        System.err.println(ex.currentToken);
//        //ex.printStackTrace();
//      }
      //System.out.println("FIN ANALYSE LEXICOSYNTAXIQUE");
      if( tmpTreeAnalysed != null)
        try {
          setExpressionTree(tmpTreeAnalysed);
        }
        catch (Exception ex) {
          System.err.println(ex.getMessage());
          System.err.println("Aborting the setting of the expression tree.");
        }
      else
        System.err.println("Erreur : tmpTreeAnalysed vaut null : cause erreur dans l'analyse");
//      System.out.println(Tree.leftCourse(exprTree));
//      System.out.println("fin du parsage");
    }
  }

  /**
   * Check each leaf of the tree and check for
   * error like the leaf is a criterion that does not exist
   * in this AdvEngineSearch.
   * This is like the sementic validation for an expression.
   * NB : To be call before lunching an AdvSearch.
   * @return true if the tree is Valid for this search.
   */
  protected boolean is_ValidateTree(){
    return is_ValidateTree(exprTree);
  }

  /**
   * This function check if the tree is a valid tree.
   * It's check if each leaf that is the name of the criterion
   * exist in this search. If criterion does not exist,
   * the function return false.
   * If an error occur during the course of the tree, the function
   * will return false after writing an error message on std err.
   * @param t the tree to check
   * @return true if t is a Valid tree .
   */
  private boolean is_ValidateTree(Tree t){
    boolean left  = true;
    boolean right = true;
    try{
      if( t.is_empty() ) return true;
      else if( t.is_leaf() ) return checkCriterion(t.getValue());
      // node
      else{
        if( t.getLeftChild() != null )
          left = is_ValidateTree(t.getLeftChild());
        if( t.getRightChild() != null )
          right = is_ValidateTree(t.getRightChild());
        return left && right;
      }
    }catch(Exception e){
      System.out.println(e.getMessage());
      System.err.println("is_ValidateTree : Error during validation(semantic) of the tree in the Advanced search : aborting");
      //e.printStackTrace();
      return false;
    }
  }

  /**
   * this evaluate the tree t for a member m.
   * This means that the function course the tree and make an evaluation
   * of the member on each leaf.
   * NB : the tree has to be lexicaly , syntaxly and sementically correct.
   * NB : the evaluation tree is made for the grammar of parser. If the grammar change,
   * this fundtion have also to change.
   * @param t the expression tree.
   * @param m the member to evaluate
   * @return true if the member is ok with all the leaf( the criterion) else false.
   * @throws NullPointerException
   * @throws Exception
   * @see AdvancedEngineSearch#setExpressionTree(Tree t) setExpressionTree
   * @see AdvSearchParser
   */
  public boolean evaluateTree(Tree t, Member m)
      throws NullPointerException, Exception{
    boolean left  = true;
    boolean right = true;
    try{
      if( t.is_empty() ) return true;
      else if( t.is_leaf() )
        return  ((Criterion)(criterionList.get(t.getValue()))).isCriterionValid(m);
      // node
      else{
        if( t.getLeftChild() != null )
          left = evaluateTree(t.getLeftChild(),m);
        if( t.getRightChild() != null )
          right = evaluateTree(t.getRightChild(),m);
        // evaluation en fonction de la racine
        // AND OR ou NOT
        if( t.getValue().compareTo("AND") == 0){
          // check some error
          if( t.getLeftChild() == null || t.getRightChild() == null){
            throw new NullPointerException("exception AND have a null child !!");
          }
          return left && right;
        }
        else if( t.getValue().compareTo("OR") == 0){
          if( t.getLeftChild() == null || t.getRightChild() == null){
            throw new NullPointerException("exception OR have a null child !!");
          }
          return left || right;
        }
        else if( t.getValue().compareTo("NOT") == 0){
          return ! (left && right);
        }
        else{
          // error
          throw new Exception("AdvancedEngineSearch : unknow keyword : "+t.getValue());
        }

      }


    }catch(Exception e){
      System.err.println("Error during validation of the tree in the Advanced search : aborting");
      System.out.println(e.getMessage());
      return false;
    }
  }

  /**
   * This function execute the search on the classifier or on the inspector( it depends
   * on the constructor call at the instanciation of this object.).
   * @return the ArrayList of Member that match with the search.
   */
  public ArrayList search(){
    ArrayList res = new ArrayList();

    if(TYPE_OF_SEARCH == 1)
      res = classifierSearch();
    else if(TYPE_OF_SEARCH == 0)
      res = inspectorSearch();
    else
      System.err.println("bad type of serach");

    return res;
  }

  /**
   * A specialised function to make a search on a classifier.
   * @return the ArrayList of Member that match with the search.
   */
  private ArrayList classifierSearch(){
    ArrayList res = new ArrayList();

    // check the constructor
    Constructor[] constr = classifier.getClassAnalysed().getDeclaredConstructors();
    checkMember(constr, res );

    // check the fields
    Field[] field = classifier.getClassAnalysed().getDeclaredFields();
    checkMember(field, res);

    // check the Methods
    Method[] method = classifier.getClassAnalysed().getDeclaredMethods();
    checkMember(method, res);

    return res;
  }

  /**
   * this procedure check if each Member of mTab is a Member that match with
   * the expression tree.( if it can be a result of a search. if yes,
   * the member that match is added to res.)
   * @param mTab the tab of Membr to check.
   * @param res the ArrayList where to add the result.
   */
  private void checkMember(Member[] mTab, ArrayList res){
    try {
      for(int i = 0 ; i < mTab.length ; i++){
        if( evaluateTree(exprTree, mTab[i] ))
          res.add(mTab[i]);
      }
    }
    catch (NullPointerException ex) {
      System.err.println(ex.getMessage());
      ex.printStackTrace();
    }
    catch (Exception ex) {
      System.err.println(ex.getMessage());
      ex.printStackTrace();
    }

  }

  /**
   * A specialised function to make a search on a Inspector.
   * @return the ArrayList of Member that match with the search.
   */
  private ArrayList inspectorSearch(){
   ArrayList res = new ArrayList();

   Iterator iter = inspector.getClassifierAnalyserIterator();
   while (iter.hasNext()) {
     ClassifierAnalyser item = (ClassifierAnalyser)iter.next();
     // bidouille pour utiliser classifierSearch !!
     classifier = item;
     res.addAll(classifierSearch());
   }

   return res;
  }
}
