package fr.unice.javinspector.core.loader;

import java.io.*;
import fr.unice.javinspector.GUI.*;
import fr.unice.javinspector.core.*;

/**
 *
 * <p>Title: JavInspector</p>
 * <p>Description: Custom class ClassLoader to load classes from file
 * not in the classpath</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universit de Nice Sophia Antipolis (UNSA)</p>
 * @author Jerme GAHIDE - Lucas CHARBIT - Gabriel ZERBIB - Xavier GALBOIS
 * @version 1.0
 */

public class FileClassLoader extends GeneralClassLoader {
  /**
   * the current classpath for loading classes
   */
  private String    classPath;
  public FileClassLoader(String classPath) {
    this.classPath = classPath;
  }


  /**
   * loads a class from its name (if it is a system class or if it has alredy benn loaded
   * @param className the name of the class to load
   * @return the loaded class or null if it failed
   */
  private Class loadFromClassName(String className){
    Class result = null;
    result = (Class)classes.get(className);
    if (result != null) {
      monitor(">> returning cached result.");
      return result;
    }

    try {
      result = super.findSystemClass(className);
      monitor(">> returning system class (in CLASSPATH).");
      return result;
    } catch (ClassNotFoundException e) {
      monitor(">> Not a system class.");
    }

    return result;
  }

  /**
   * an overide of the loadClass method to do try to guess the classpath of
   * the class to load in order to load its depandencies
   * @param className the path of a file or the name of a class
   * @param resolveIt call resolveClass()
   * @return the loaded class
   * @throws ClassNotFoundException
   */
  public synchronized Class loadClass(String className,
                                       boolean resolveIt) throws ClassNotFoundException {
     String realClassName;
     Class   result = null;
     byte[]  classBytes;
     monitor(">> GeneralClassLoader.loadClass(" + className + ", " + resolveIt + ")");

     if ((className.indexOf(File.separatorChar) < 0) && !(className.endsWith(".class"))){
       // the classname is a 'real' classname i.e. java.lang.String
       realClassName = className;
       result = loadFromClassName(className);

       if (result == null) {
         classBytes = loadClassBytes(realClassName);
         if (classBytes == null) {
           throw new ClassNotFoundException();
           //return null;
         }
        try {
           result = defineClass(null, classBytes, 0, classBytes.length);
        }

        catch (Throwable e) {
          result = null;
        }
       }
       if (result != null) {
         if (resolveIt)
           resolveClass(result);
         classes.put(result.getName(), result);
         monitor(">> caching result : " + result.getName());
       }
     }
     else {
       // the classname is a file path
       int i = 0;
       while  ((realClassName = guessClassName(className, i)) != null){
         i++;
         result = loadFromClassName(realClassName);
         if (result != null) return result;

         classBytes = loadClassBytes(realClassName);
         if (classBytes == null) {
           continue;
         }

         try {
           result = defineClass(null, classBytes, 0, classBytes.length);
         }
         catch (Throwable e) {
           monitor("### FAILED ==> RETRYING ###");
           continue;
         }
         if (result != null) {
           prepareResolve(className, result.getName());
           if (resolveIt) resolveClass(result);
           Class tmp;
           if (( tmp = loadFromClassName(result.getName())) != null){
             result = tmp;
           }
           else {
             classes.put(result.getName(), result);
             monitor(">> caching result : " + result.getName());
           }
           return result;
         }
       }
     }
     if (result == null) throw new ClassNotFoundException("could not load the specified class : " + className);
     return result;


 }
  /**
   * loads data from a file (supposed to be a .class)
   * @param className full name of the class
   * @return the bytes of the file
   */
  protected byte[] loadClassBytes(String className) {
    monitor(className);
    setClassNameReplacementChar(File.separatorChar);
    className = formatClassName(className);
    monitor(">> from file: " + classPath + className);
    byte result[];
    String fileName = classPath + className;
    try {
      FileInputStream inStream = new FileInputStream(fileName);
      result = new byte[inStream.available()];
      inStream.read(result);
      inStream.close();
      return result;
    }
    catch (Exception e) {
      return null;
    }
  }
  public void setFilePrefix(String str){
    classPath = str;
  }

  /**
   * set the classpath and return the name of class from its path
   * in order to load its dependencies
   * @param filePath the path
   * @param i the number of directories to go up
   * @return the full name of the class
   */
  public String guessClassName(String filePath, int i){
    int indexRes = filePath.length();
    String res;
    int cur = 0;
    while (cur <= i){
      indexRes = filePath.lastIndexOf(File.separatorChar, indexRes - 1);
      if (indexRes < 0) return null;
      cur++;
    }
    this.setFilePrefix(filePath.substring(0, indexRes + 1));
    monitor(">>> prefix set to : " + classPath);
    res = filePath.substring(indexRes + 1);
    res = res.replace(File.separatorChar, '.');
    if (res.endsWith(".class")) res = res.substring(0, res.length() - 6);
    monitor(">>> guessed class name : " + res);
    return res;
  }

  /**
   * Sets the class path from a the full path and the full name of
   * a Class already loaded
   * @param name the full path of the .class file from which it was loaded
   * @param fullClassName the full name of the class
   */
  protected void prepareResolve(String name, String fullClassName){
    String tmpFilePrefix;
    String pkgName;
    String className;
    int lastSlashIndex = name.lastIndexOf(File.separatorChar);
    int lastDotIndex;
    int lastSpaceIndex;
    if (lastSlashIndex < 0) {
        monitor("### erreur chaine innatendue");
        return;
    }
    tmpFilePrefix = name.substring(0, lastSlashIndex);
    if ((lastDotIndex = fullClassName.lastIndexOf('.')) > 0)
      pkgName = fullClassName.substring(0, lastDotIndex);
    else pkgName = "";

    className = fullClassName.substring(lastDotIndex + 1);
    if (pkgName.length() > tmpFilePrefix.length()) return;
    classPath = tmpFilePrefix.substring(0, tmpFilePrefix.length() - pkgName.length());
    setClassNameReplacementChar(File.separatorChar);

    monitor("-->" + classPath);
    monitor("-->" + tmpFilePrefix);
    monitor("--> " + fullClassName);
    monitor("--> " + pkgName);
    monitor("--> " + className);

  }
}


