package fr.unice.javinspector.core.loader;


import java.util.Hashtable;
import fr.unice.javinspector.GUI.*;
import java.io.File;
import java.io.*;


/**
 *
 * <p>Title: JavInspector</p>
 * <p>Description: a General Abstrract Class ClassLoader</p>
 * <p>Company: Universit de Nice Sophia Antipolis (UNSA)</p>
 * @author Jerme GAHIDE - Lucas CHARBIT - Gabriel ZERBIB - Xavier GALBOIS
 * @version 1.0
 */
public abstract class GeneralClassLoader extends ClassLoader {
  /**
   * Cache of already loaded classes
   */
  protected static Hashtable classes = new Hashtable();
  /**
   * the package separator replacement char
   */
  private char      classNameReplacementChar = 0;
  /**
   * sets verbose monitor on or off
   */
  protected boolean   monitorOn = false;
  protected boolean   sourceMonitorOn = false;

  /**
   * Default constructor
   */
  public GeneralClassLoader() {
  }

  /**
   * loads the class <code>className</code>
   * this call is equivalent to : <code>loadClass(className, true);</code>
   * @param className the name of the class to load
   * @return the loaded class if it succeed
   * @throws ClassNotFoundException
   */
  public Class loadClass(String className) throws ClassNotFoundException {
    return (loadClass(className, true));
  }


  /**
   * loads the class <code>className</code>
   * @param className the name of the class to load
   * @param resolveIt tells wether the ClassLoader should call the :
   * <code>parent.resolvClass(className);</code>
   * @return the loaded class if it succeed
   * @throws ClassNotFoundException
   * @throws NoClassDefFoundError
   */
  public synchronized Class loadClass(String className,
          boolean resolveIt) throws ClassNotFoundException {

      Class   result;
      byte[]  classBytes;
      monitor(">> MultiClassLoader.loadClass(" + className + ", " + resolveIt + ")");

      //----- Check our local cache of classes
      result = (Class)classes.get(className);
      if (result != null) {
          monitor(">> returning cached result.");
          return result;
      }

      //----- Check with the primordial class ClassLoader
      try {
          result = super.findSystemClass(className);
          monitor(">> returning system class (in CLASSPATH).");
          return result;
      } catch (ClassNotFoundException e) {
        monitor(">> Not a system class.");

        //----- Try to load it from preferred source
        // Note loadClassBytes() is an abstract method
        classBytes = loadClassBytes(className);
        if (classBytes == null) {
          monitor(">> could not load Class Bytes");
          throw new ClassNotFoundException(e.getMessage());
        }

        //----- Define it (parse the class file)
        result = defineClass(null, classBytes, 0, classBytes.length);
        if (result == null) {
          monitor(">> could not define Class");
          throw new ClassFormatError("File is in an unknow format");
        }

      }
      //----- Resolve if necessary
      if (resolveIt) resolveClass(result);

      // Done
      classes.put(result.getName(), result);
      monitor(">> caching result : " + result.getName());
      return result;
  }

  /**
   * Sets the package spearator char (default is '.')
   * @param replacement
   */
  public void setClassNameReplacementChar(char replacement) {
    classNameReplacementChar = replacement;
  }

  /**
   * Abstract implentation
   * needs to be defined in the subclasses
   * @param className a class name
   * @return the byte array containing the byte code
   */
  protected abstract byte[] loadClassBytes(String className);


//  protected abstract void prepareResolve(String name, String fullClassName);
//
//  public abstract String guessClassName(String filePath, int i);


  /**
   * Changes the prefix of the class to load (ex : the classpath or the jar file)
   * @param str
   */
  public abstract void setFilePrefix(String str) throws IOException;

  /**
   * formats a classname to a file name
   * it uses the classNameReplacementChar
   * @see #setClassNameReplacementChar
   * @param className
   * @return the formated string
   */
  protected String formatClassName(String className) {
    String res;
    if (className.endsWith(".class")) className = className.substring(0, className.length() - 6);
    if (classNameReplacementChar == 0) {
      res = className;
    } else {
      res = className.replace('.', classNameReplacementChar);
    }
    return res + ".class";
  }

  protected void monitor(String text) {
    if (monitorOn) print(text);
  }

  protected static void print(String text) {
    System.out.println(text);
  }
}
